# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import functools

import six

import sqlalchemy as sa
import sqlalchemy.orm  # noqa
from sqlalchemy.ext.declarative import declarative_base, declared_attr

from django.db import DEFAULT_DB_ALIAS

from ..utils import make_args, setdefaultattr, suppress
from .models import Base
from .query import Query, QueryProperty
from .session import SignallingSession
from .signals import all_signals, engine_created
from .url import make_url


class TransactionContext(object):
    """
    Transaction context manager for maintaining a transaction or savepoint
    """

    def __init__(self, db, savepoint=True):
        self.db = db
        self.savepoint = savepoint
        self.transaction = None

    def __call__(self, func, *args, **kwargs):

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            with self.db.begin(subtransactions=True, nested=self.savepoint):
                return func(*args, **kwargs)

        return wrapped

    def __enter__(self):
        self.transaction = self.db.begin(subtransactions=True, nested=self.savepoint)
        self.transaction.__enter__()

    def __exit__(self, exception_type, value, traceback):
        self.transaction.__exit__(exception_type, value, traceback)
        self.transaction = None


class _sqla_meta(type):

    def __new__(cls, name, bases, attrs):
        typ = super(_sqla_meta, cls).__new__(cls, name, bases, attrs)

        # figure out all props to be proxied
        props = set([i for i in dir(sa.orm.Session()) if not i.startswith("__")])
        props.update(sa.orm.Session.public_methods)

        for i in props:
            if not hasattr(typ, name):
                if hasattr(sa.orm.Session, i) and callable(getattr(sa.orm.Session, i)):
                    setattr(typ, i, sa.orm.scoping.instrument(i))
                else:
                    setattr(typ, i, sa.orm.scoping.makeprop(i))

        return typ


class SQLAlchemy(six.with_metaclass(_sqla_meta, object)):
    """
    This class itself is a scoped session and provides very thin and useful abstractions and conventions for using
    sqlalchemy with django.
    """
    session_class = SignallingSession
    query_class = Query
    registry_class = sa.util.ThreadLocalRegistry
    model_class = Base

    def __init__(self, alias=None, **kwargs):
        self.alias = alias or DEFAULT_DB_ALIAS
        self.url, self.kwargs = self._make_url(alias)
        self.kwargs.update(kwargs)
        self.session_class = kwargs.pop("session_class", None) or self.session_class
        self.query_class = kwargs.pop("query_class", None) or self.query_class
        self.registry_class = kwargs.pop("registry_class", None) or self.registry_class
        self.model_class = kwargs.pop("model_class", None) or self.model_class
        self.engine_options = kwargs.pop("engine_options", {})
        self.session_options = kwargs.pop("session_options", {})

        self.session_options.setdefault("query_cls", self.query_class)
        self.session_options.setdefault("class_", self.session_class)

        self.middleware = self.make_middleware()
        self.Model = self._make_declarative(self.model_class)

        for module in sa, sa.sql, sa.orm:
            for key in module.__all__:
                if not hasattr(self, key):
                    setattr(self, key, getattr(module, key))

        self.collections = sa.orm.collections
        self.event = sa.event
        self.relationship = self._wrap(self.relationship)
        self.relation = self._wrap(self.relation)
        self.dynamic_loader = self._wrap(self.dynamic_loader)

    def __call__(self, **kwargs):
        return self.session(**kwargs)

    @property
    def registry(self):
        if not hasattr(self, "_registry"):
            engine = self._create_engine(self.url, **self.engine_options)
            self._registry = self.registry_class(sa.orm.sessionmaker(bind=engine, **self.session_options))

        return self._registry

    def session(self, **kwargs):
        """
        Return the current session, creating it if necessary using session_factory for the current scope
        Any kwargs provided will be passed on to the session_factory. If there's already a session in
        current scope, will raise InvalidRequestError
        """
        if kwargs:
            if self.registry.has():
                raise sa.exc.InvalidRequestError(
                    "Scoped session is already present; " "no new arguments may be specified."
                )

            else:
                session = self.session_factory(**kwargs)
                self.registry.set(session)
                return session

        else:
            return self.registry()

    def _get_kwargs_for_relation(self, kwargs, prefix="fk_"):
        opts = {}
        for key in list(kwargs.keys()):
            if key.startswith(prefix):
                opts[key] = kwargs.pop(key)
        return opts

    def OneToMany(self, remote_cls, **kwargs):
        """
        Use an event to build one-to-many relationship on a model and auto generates foreign key relationship from the
        remote table::

            class ModelOne(db.Model):
                pk = db.Column(.., primary_key=True)
                m2 = db.OneToMany("ModelTwo", ...)

            class ModelTwo(db.Model):
                pk = db.Column(.., primary_key=True)
                ...

            will create ModelTwo.m1_pk automatically for the relationship
        """

        @declared_attr
        def o2m(cls):
            """
            one to many relationship attribute for declarative
            """
            rels = setdefaultattr(cls, "_relationships", set())
            kwargs.setdefault("info", {}).update(self._get_kwargs_for_relation(kwargs))
            kwargs["uselist"] = True
            backref = kwargs.get("backref")
            backref_kwargs = None
            if backref:
                if isinstance(backref, tuple):
                    with suppress(Exception):
                        backref, backref_kwargs = backref

                backref_kwargs = backref_kwargs or {}

                backref_kwargs["uselist"] = False
                kwargs["backref"] = (backref, backref_kwargs)

            rel = self.relationship(remote_cls, **kwargs)
            rel.direction = sa.orm.interfaces.ONETOMANY
            rels.add(rel)
            return rel

        return o2m

    def ManyToOne(self, remote_cls, **kwargs):
        """
        Use an event to build many-to-one relationship on a model and auto generates foreign key relationship on the
        remote table::

            class ModelOne(db.Model):
                pk = db.Column(.., primary_key=True)
                m2 = db.ManyToOne("ModelTwo", ...)

            class ModelTwo(db.Model):
                pk = db.Column(.., primary_key=True)
                ...

        will create ModelOne.m2_pk automatically for the relationship
        """

        @declared_attr
        def m2o(cls):
            """
            many to one relationship attribute for declarative
            """
            rels = setdefaultattr(cls, "_relationships", set())
            kwargs.setdefault("info", {}).update(self._get_kwargs_for_relation(kwargs))
            kwargs["uselist"] = False
            backref = kwargs.get("backref")
            if backref:
                backref_kwargs = None
                if isinstance(backref, tuple):
                    with suppress(Exception):
                        backref, backref_kwargs = backref

                backref_kwargs = backref_kwargs or {}

                backref_kwargs["uselist"] = True
                kwargs["backref"] = self.backref(backref, **backref_kwargs)

            rel = self.relationship(remote_cls, **kwargs)
            rel.direction = sa.orm.interfaces.MANYTOONE
            rels.add(rel)
            return rel

        return m2o

    def ManyToMany(self, remote_cls, table_name=None, **kwargs):
        """
        Use an event to build many-to-many relationship on a model and auto generates an association table or if a
        model is provided as secondary argument::

            class ModelOne(db.Model):
                pk = db.Column(.., primary_key=True)
                m2s = db.ManyToMany("ModelTwo", backref="m1s", table_name='m1m2s', ...)

            class ModelTwo(db.Model):
                pk = db.Column(.., primary_key=True)
                ...

        or with back_populates::

            class ModelOne(db.Model):
                pk = db.Column(.., primary_key=True)
                m2s = db.ManyToMany("ModelTwo", back_populates="m1s", table_name='m1m2s', ...)

            class ModelTwo(db.Model):
                pk = db.Column(.., primary_key=True)
                m1s = db.ManyToMany("ModelOne", back_populates="m2s", table_name='m1m2s', ...)

        will create ModelOne.m2s and ModelTwo.m1s relationship thru a provided secondary argument. If no secondary argument
        is provided, table_name is required as it will be used for the autogenerated association table.

        In the case of back_populates you have to provide the same table_name argument on both many-to-many
        declarations
        """

        @declared_attr
        def m2m(cls):
            """
            many to many relationship attribute for declarative
            """
            if "secondary" not in kwargs and table_name is None:
                raise sa.exc.ArgumentError(
                    "You need to provide secondary or table_name for the relation for the association table "
                    "that will be generated"
                )

            rels = setdefaultattr(cls, "_relationships", set())
            info = kwargs.setdefault("info", {})
            info.update(self._get_kwargs_for_relation(kwargs))
            info.update(self._get_kwargs_for_relation(kwargs, "table_"))
            if table_name:
                info["table_name"] = table_name

            kwargs["uselist"] = True

            backref = kwargs.get("backref")
            backref_kwargs = None
            if backref:
                if isinstance(backref, tuple):
                    with suppress(Exception):
                        backref, backref_kwargs = backref

                backref_kwargs = backref_kwargs or {}

                backref_kwargs["uselist"] = True
                kwargs["backref"] = self.backref(backref, **backref_kwargs)

            rel = self.relationship(remote_cls, **kwargs)
            rel.direction = sa.orm.interfaces.MANYTOMANY
            rels.add(rel)
            return rel

        return m2m

    def Table(self, name, *args, **kwargs):
        assert name is not None, "Table requires `name` argument"
        assert len(args) > 0, "Table at least takes one column argument"
        for arg in args:
            assert not isinstance(arg, sa.MetaData), "Passing a `metadata` is not allowed"

        return sa.Table(name, self.metadata, *args, **kwargs)

    def _set_query_class(self, cls, kwargs):
        query_class = self.query_class
        try:
            query_class = cls.query_class
        except AttributeError:
            pass

        kwargs["query_class"] = query_class

    def _wrap(self, fn):
        """
        Wrapper that sets the 'query_class' argument from the model's
        query_class attribute
        """

        @functools.wraps(fn)
        def func(cls, *args, **kwargs):
            """
            Wraps a function setting default `query_class` argument
            """
            self._set_query_class(cls, kwargs)

            if "backref" in kwargs:
                backref = kwargs["backref"]
                backref_kwargs = {}
                if isinstance(backref, tuple):
                    backref, backref_kwargs = backref
                    self._set_query_class(backref, backref_kwargs)

                kwargs["backref"] = (backref, backref_kwargs)

            return fn(cls, *args, **kwargs)

        return func

    def _create_engine(self, url, **kwargs):
        engine = sa.create_engine(url, **kwargs)
        engine_created.send(engine)
        return engine

    def _make_url(self, alias):
        """
        Makes a url out of alias or url string
        ----------------------------------------------------------
        alias: str
            a config alias or url string
        """
        return make_url(alias)

    def _make_declarative(self, model):
        """
        Creates the base class that the models should inherit from
        ----------------------------------------------------------
        model: class
            The base class for the declarative_base to be inherited from
        """
        base = declarative_base(cls=model)

        # allow to customize things in custom base model
        if not hasattr(base, "query_class"):
            base.query_class = self.query_class
        if not hasattr(base, "query"):
            base.query = self.queryproperty()
        if not hasattr(base, "objects"):
            base.objects = self.queryproperty()

        return base

    @property
    def engine(self):
        """
        Current engine
        """
        return self.bind

    @property
    def metadata(self):
        """
        Current metadata
        """
        return self.Model.metadata

    @property
    def session_factory(self):
        """
        Current session factory to create sessions
        """
        return self.registry.createfunc

    def __repr__(self):
        return "<{} engine={!r}>".format(self.__class__.__name__, self.url)

    def queryproperty(self, *args, **kwargs):
        """
        Generate a query property for a model
        """
        model = None
        if args and isinstance(args[0], type) and issubclass(args[0], self.Model):
            model = args[0]
            args = args[1:]

        return QueryProperty(self, model, *args, **kwargs)

    def atomic(self, savepoint=True):
        """
        Create a savepoint or transaction scope
        """
        return TransactionContext(self, savepoint)

    def make_middleware(self):
        """
        Creates a middleware to be used in a django application
        """
        from .middleware import SQLAlchemyDBMiddleware

        return type(str("{}SQLAlchemyMiddleware".format(self.alias)), (SQLAlchemyDBMiddleware,), {"db": self})

    def args(self, *args, **kwargs):
        """
        Useful for setting table args and mapper args on models
        """
        return make_args(*args, **kwargs)

    def remove(self):
        """
        Remove the current scoped session
        """
        if self.registry.has():
            self.registry().close()
        self.registry.clear()
        for signal in all_signals.scoped_signals:
            signal.cleanup()

    def create_all(self):
        """
        Create the schema in db
        """
        self.metadata.create_all(bind=self.engine)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from collections import OrderedDict
from itertools import chain

import six

import sqlalchemy as sa
import sqlalchemy.ext.declarative  # noqa
import sqlalchemy.orm  # noqa
from sqlalchemy.orm.base import NO_VALUE

from .meta import model_info


def get_primary_keys(model, kwargs):
    """
    Returns the primary key from a dictionary to be used in a sqlalchemy query.get() call
    """
    info = model_info(model)
    pks = []

    for attr, _ in info.primary_keys.items():
        pk = kwargs.get(attr)
        pks.append(pk)

    if len(pks) < 2:
        return next(iter(pks), None)

    if any(pk is None for pk in pks):
        return None

    return tuple(pks)


def get_primary_keys_from_instance(instance):
    """
    Return a dict containing the primary keys of the ``instance``
    """
    if instance is None:
        return None

    info = model_info(type(instance))

    if len(info.primary_keys) > 1:
        return OrderedDict((name, getattr(instance, name)) for name in info.primary_keys)

    return getattr(instance, next(iter(info.primary_keys)))


def model_to_dict(instance, fields=None, exclude=None):
    """
    Return a dict containing the data in ``instance`` suitable for passing as
    a Form's ``initial`` keyword argument.

    ``fields`` is an optional list of field names. If provided, return only the
    named.

    ``exclude`` is an optional list of field names. If provided, exclude the
    named from the returned dict, even if they are listed in the ``fields``
    argument.
    """
    info = model_info(type(instance))

    fields = set(
        fields or list(info.properties.keys()) + list(info.primary_keys.keys()) + list(info.relationships.keys())
    )
    exclude = set(exclude or [])
    data = {}
    for name in info.properties:

        if name.startswith("_"):
            continue

        if name not in fields:
            continue

        if name in exclude:
            continue

        data[name] = getattr(instance, name)

    for name, rel in info.relationships.items():

        if name.startswith("_"):
            continue

        if name not in fields:
            continue

        if name in exclude:
            continue

        if rel.uselist:
            for obj in getattr(instance, name):
                pks = get_primary_keys_from_instance(obj)
                if pks:
                    data.setdefault(name, []).append(pks)
        else:
            obj = getattr(instance, name)
            pks = get_primary_keys_from_instance(obj)
            if pks:
                data[name] = pks

    return data


def simple_repr(instance, fields=None):
    """
    Provides basic repr for models
    ------------------------------
    instance: Model
        a model instance
    fields: list
        list of fields to display on repr
    """
    state = sa.inspect(instance)
    pks = [state.mapper.get_property_by_column(col).key for col in state.mapper.primary_key]
    fields = fields or [
        state.mapper.get_property_by_column(col).key for col in state.mapper.columns if not col.primary_key
    ]

    pk_reprs = []
    for key in pks:
        value = state.attrs[key].loaded_value
        if value == NO_VALUE:
            value = None
        elif isinstance(value, six.text_type):
            # remove pesky "u" prefix in repr for strings
            value = str(value)
        pk_reprs.append("=".join([key, repr(value)]))

    field_reprs = []
    for key in fields:
        value = state.attrs[key].loaded_value
        if isinstance(value, six.text_type):
            # remove pesky "u" prefix in repr for strings
            value = str(value)
        if value != NO_VALUE:
            field_reprs.append("=".join([key, repr(value)]))

    return "".join([instance.__class__.__name__, "(", ", ".join(chain(pk_reprs, sorted(field_reprs))), ")"])


def serialize(instance, *rels):
    """
    Return a dict of column attributes
    ------------------------------
    instance:
        a model instance
    rels: list of relations
        relationships to be serialized
    """
    if instance is None:
        return None

    if isinstance(instance, (list, set)):
        return [serialize(i, *rels) for i in instance]

    state = sa.inspect(instance)

    data = {c.key: getattr(instance, c.key) for c in state.mapper.column_attrs}

    rels = set(rels)

    for relation in state.mapper.relationships:
        attr = getattr(state.mapper.class_, relation.key)
        if attr in rels:
            sub_instance = getattr(instance, relation.key, None)
            sub_rels = [r for r in rels if r is not attr]
            data[relation.key] = serialize(sub_instance, *sub_rels)

    return data


def clone(instance, *rels, **kwargs):
    """
    Clone a model instance with or without any relationships
    --------------------------------------------------------
    instance: Model
        a model instance
    relations: list or relations or a tuple of relation and kwargs for that relation
        relationships to be cloned with relationship and optionally kwargs
    kwargs: dict string of any
        attribute values to be overridden
    """
    if instance is None:
        return None

    if isinstance(instance, (list, set)):
        return [clone(i, *rels, **kwargs) for i in instance]

    relations = {}
    for rel in rels:
        r = rel
        rkwargs = {}
        if isinstance(rel, tuple):
            r, rkwargs = rel

        relations[r] = rkwargs

    state = sa.inspect(instance)
    mapper = state.mapper
    kwargs = kwargs or {}

    fks = set(
        chain(
            *[
                fk.columns
                for fk in chain(*[table.constraints for table in mapper.tables])
                if isinstance(fk, sa.ForeignKeyConstraint)
            ]
        )
    )

    for column in mapper.columns:
        prop = mapper.get_property_by_column(column)

        if prop.key in kwargs:
            continue

        if column.primary_key or column in fks:
            continue

        kwargs[prop.key] = getattr(instance, prop.key)

    for relation in mapper.relationships:
        if relation.key in kwargs:
            continue

        attr = getattr(mapper.class_, relation.key)
        if attr in relations:
            sub_instance = getattr(instance, relation.key, None)
            sub_rels = [(r, kw) for r, kw in relations.items() if r is not attr]
            kwargs[relation.key] = clone(sub_instance, *sub_rels, **relations[attr])

    cloned = mapper.class_(**kwargs)
    return cloned


class Base(object):
    """
    Base model class for SQLAlchemy.

    Can be overwritten by subclasses:

        query_class = None

    Automatically added by declarative base for easier querying:

        query = None
        objects = None
    """

    @sa.ext.declarative.declared_attr
    def __tablename__(cls):
        """
        Autogenerate a table name
        """
        return cls.__name__.lower()

    def as_dict(self):
        """
        Return a dict of column attributes
        """
        return serialize(self)

    def __repr__(self):
        return simple_repr(self)

    @classmethod
    def __declare_first__(cls):
        """
        Declarative hook called within `before_configure` mapper event, can be called multiple times.
        """
        rels = getattr(cls, "_relationships", set())
        for relation in rels:
            if relation.direction == sa.orm.interfaces.ONETOMANY:
                relation.mapper.class_._add_foreign_keys(cls, relation)
            elif relation.direction == sa.orm.interfaces.MANYTOONE:
                cls._add_foreign_keys(relation.mapper.class_, relation)
            elif relation.direction == sa.orm.interfaces.MANYTOMANY:
                cls._add_association_table(relation.mapper.class_, relation)

        rels.clear()

    @classmethod
    def _add_foreign_keys(cls, parent_cls, relation):
        """
        Generate fk columns and constraint to the remote class from a relationship
        """
        fk_kwargs = {key[3:]: val for key, val in relation.info.items() if key.startswith("fk_")}
        fk_prefix = fk_kwargs.pop("fk_prefix", "_")
        fk_nullable = fk_kwargs.pop("fk_nullable", True)
        fk_key = fk_kwargs.pop("fk_key", None)

        if not fk_key:
            if relation.direction == sa.orm.interfaces.MANYTOONE:
                fk_key = relation.key.lower()
            elif relation.backref:
                backref, _ = relation.backref
                fk_key = backref.lower()
            else:
                fk_key = parent_cls.__name__.lower()

        cols = {}
        cols_created = False
        for pk_column in parent_cls.__table__.primary_key:
            pk_attr = parent_cls.__mapper__.get_property_by_column(pk_column)
            col_name = "_".join(filter(None, [fk_key, pk_column.name]))
            attr = "{}{}".format(fk_prefix, "_".join(filter(None, [fk_key, pk_attr.key])))

            if col_name not in cls.__table__.columns and not hasattr(cls, attr):
                fk_column = sa.Column(col_name, pk_column.type, nullable=fk_nullable)
                setattr(cls, attr, fk_column)
                cols_created = True
            else:
                fk_column = cls.__table__.columns[col_name]

            cols[pk_column] = fk_column

        relation._user_defined_foreign_keys = cols.values()

        if cols_created:
            # pk and fk ordering must match for foreign key constraint
            pks, fks = [], []
            for pk in cols:
                pks.append(pk)
                fks.append(cols[pk])

            constraint = sa.ForeignKeyConstraint(fks, pks, **fk_kwargs)
            cls.__table__.append_constraint(constraint)

    @classmethod
    def _add_association_table(cls, child_cls, relation):
        """
        Generate association table and fk constraints to satisfy a many-to-many relation
        """
        if relation.secondary is not None:
            return

        table_name = relation.info.get("table_name")
        relation.secondary = cls.metadata.tables.get(table_name)
        if relation.secondary is not None:
            return

        fk_kwargs = {key[3:]: val for key, val in relation.info.items() if key.startswith("fk_")}
        table_kwargs = {key[6:]: val for key, val in relation.info.items() if key.startswith("table_")}
        table_kwargs.pop("name", None)

        column_map = {}
        for pk_column in chain(cls.__mapper__.primary_key, child_cls.__table__.primary_key):
            col_name = "_".join(filter(None, [pk_column.table.name.lower(), pk_column.name]))
            col = sa.Column(col_name, pk_column.type, primary_key=True)
            column_map.setdefault(pk_column.table, []).append(col)

        table_args = list(chain(*column_map.values()))

        for table, columns in column_map.items():
            table_args.append(sa.ForeignKeyConstraint(columns, table.primary_key, **fk_kwargs))

        relation.secondary = sa.Table(
            table_name, cls.metadata, *table_args, schema=cls.__table__.schema, **table_kwargs
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import contextlib
import inspect


try:
    suppress = contextlib.suppress
except AttributeError:  # pragma: nocover
    from .compat import suppress  # noqa pragma: nocover


def get_args(func):
    """
    Returns the names of the positional arguments for composite model inspection
    """
    try:
        return list(inspect.signature(func).parameters.keys())[1:]

    except AttributeError:  # pragma: nocover
        return inspect.getargspec(func).args[1:]  # pylint:disable=deprecated-pragma pragma:nocover


def setdefaultattr(obj, name, value):
    """
    setdefault for object attributes
    """
    try:
        return getattr(obj, name)

    except AttributeError:
        setattr(obj, name, value)
    return getattr(obj, name)


def make_args(*args, **kwargs):
    """
    Useful for setting table args and mapper args on models and other things
    """
    return tuple(args) + (kwargs,)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import unittest

from django.conf import settings

from django_sorcery.db.sqlalchemy import SQLAlchemy
from django_sorcery.db.utils import dbdict

from ..base import mock


class TestDbDict(unittest.TestCase):

    def setUp(self):
        super(TestDbDict, self).setUp()
        self.settings = mock.patch.dict(settings.SQLALCHEMY_CONNECTIONS, {"minimal": {"DIALECT": "sqlite"}})
        self.settings.start()
        self.addCleanup(self.settings.stop)

    def test_get(self):
        databases = dbdict()

        db = databases.get("minimal")
        self.assertEqual(str(db.url), "sqlite://")

    def test_get_again(self):
        databases = dbdict()

        db1 = databases.get("minimal")

        db2 = databases.get("minimal")
        self.assertEqual(db1, db2)

    def test_update(self):
        databases = dbdict()

        databases.update({"one": SQLAlchemy("sqlite://")}, two=SQLAlchemy("sqlite://"))

    def test_setitem(self):
        databases = dbdict()
        databases["one"] = SQLAlchemy("sqlite://")

    def test_setitem_again(self):
        databases = dbdict()
        databases["one"] = SQLAlchemy("sqlite://")

        with self.assertRaises(RuntimeError):
            databases["one"] = SQLAlchemy("sqlite://")

    def test_setitem_again_same_db(self):
        databases = dbdict()

        db = SQLAlchemy("sqlite://")
        databases["one"] = db

        with self.assertRaises(RuntimeError):
            databases["two"] = db

    def test_setitem_dummy(self):
        databases = dbdict()

        with self.assertRaises(RuntimeError):
            databases["one"] = {}



# -*- coding: utf-8 -*-
"""
Field mapping from SQLAlchemy type's to form fields
"""
from __future__ import absolute_import, print_function, unicode_literals
import datetime
import decimal
from collections import OrderedDict
from itertools import chain

import six

import sqlalchemy as sa

from django import forms as djangoforms
from django.forms import fields as djangofields

from .db import meta
from .fields import EnumField, ModelChoiceField, ModelMultipleChoiceField, apply_limit_choices_to_form_field


ALL_FIELDS = "__all__"

FIELD_LOOKUP = {
    # only python types here
    bool: djangofields.BooleanField,
    datetime.date: djangofields.DateField,
    datetime.datetime: djangofields.DateTimeField,
    datetime.time: djangofields.TimeField,
    datetime.timedelta: djangofields.DurationField,
    decimal.Decimal: djangofields.DecimalField,
    float: djangofields.FloatField,
    int: djangofields.IntegerField,
    str: djangofields.CharField,
}
FIELD_LOOKUP.update({t: djangofields.CharField for t in six.string_types})
FIELD_LOOKUP.update({t: djangofields.IntegerField for t in six.integer_types})


class ModelFieldMapper(OrderedDict):

    def __init__(self, opts, formfield_callback=None, apply_limit_choices_to=True):
        self.opts = opts
        self.formfield_callback = formfield_callback
        self.apply_limit_choices_to = apply_limit_choices_to

    def get_default_kwargs(self, name, **kwargs):
        if self.opts.widgets and name in self.opts.widgets:
            kwargs["widget"] = self.opts.widgets[name]
        if self.opts.localized_fields == ALL_FIELDS:
            kwargs["localize"] = True
        if self.opts.localized_fields and name in self.opts.localized_fields:
            kwargs["localize"] = True
        if self.opts.labels and name in self.opts.labels:
            kwargs["label"] = self.opts.labels[name]
        if self.opts.help_texts and name in self.opts.help_texts:
            kwargs["help_text"] = self.opts.help_texts[name]
        if self.opts.error_messages and name in self.opts.error_messages:
            kwargs["error_messages"] = self.opts.error_messages[name]
        if self.opts.field_classes and name in self.opts.field_classes:
            kwargs["form_class"] = self.opts.field_classes[name]
        return kwargs

    def get_fields(self):

        field_list = []

        info = meta.model_info(self.opts.model)

        for name, attr in chain(info.properties.items(), info.relationships.items()):

            if name.startswith("_"):
                continue

            if self.opts.fields and name not in self.opts.fields:
                continue

            if self.opts.exclude and name in self.opts.exclude:
                continue

            kwargs = self.get_default_kwargs(name)
            formfield = self.build_field(attr, **kwargs)

            if formfield:
                if self.apply_limit_choices_to:
                    apply_limit_choices_to_form_field(formfield)
                field_list.append((name, formfield))

        return OrderedDict(field_list)

    def build_field(self, info, **kwargs):
        if self.formfield_callback is not None:
            return self.formfield_callback(info, **kwargs)

        if isinstance(info, meta.column_info):
            return self.build_standard_field(info, **kwargs)

        if isinstance(info, meta.relation_info):
            return self.build_relationship_field(info, **kwargs)

    def build_relationship_field(self, relation, **kwargs):
        if relation.direction == sa.orm.relationships.MANYTOMANY:
            return self.build_modelmultiplechoice_field(relation, **kwargs)

        if relation.direction == sa.orm.relationships.ONETOMANY and relation.uselist:
            return self.build_modelmultiplechoice_field(relation, **kwargs)

        return self.build_modelchoice_field(relation, **kwargs)

    def build_modelchoice_field(self, relation, **kwargs):
        kwargs["required"] = all([col.nullable for col in relation.foreign_keys])
        return ModelChoiceField(relation.related_model, self.opts.session, **kwargs)

    def build_modelmultiplechoice_field(self, relation, **kwargs):
        kwargs["required"] = False
        return ModelMultipleChoiceField(relation.related_model, self.opts.session, **kwargs)

    def build_standard_field(self, attr, **kwargs):

        for base in attr.column.type.__class__.mro():
            type_func = getattr(self, "build_{}_field".format(base.__name__.lower()), None)
            if type_func and callable(type_func):
                kwargs.update(attr.field_kwargs)
                return type_func(attr, **kwargs)

        for base in attr.column.type.python_type.mro():
            if base in FIELD_LOOKUP:
                kwargs.update(attr.field_kwargs)
                return FIELD_LOOKUP[base](**kwargs)

    def build_enum_field(self, attr, **kwargs):
        kwargs.pop("max_length", None)

        if attr.column.type.enum_class:
            kwargs["enum_class"] = attr.column.type.enum_class
            return EnumField(**kwargs)

        kwargs["choices"] = [(val, val) for val in attr.column.type.enums]
        return djangofields.TypedChoiceField(**kwargs)

    def build_boolean_field(self, attr, **kwargs):
        if attr.column.nullable:
            return djangofields.NullBooleanField(**kwargs)

        return djangofields.BooleanField(**kwargs)

    def build_integer_field(self, attr, **kwargs):
        return djangofields.IntegerField(**kwargs)

    build_integer_field = build_integer_field
    build_smallinteger_field = build_integer_field
    build_biginteger_field = build_integer_field

    def build_text_field(self, attr, **kwargs):
        kwargs["widget"] = djangoforms.Textarea
        return djangofields.CharField(**kwargs)

    build_clob_field = build_text_field

    def build_binary_field(self, attr, **kwargs):
        pass

    build_blob_field = build_binary_field
    build_largebinary_field = build_binary_field
    build_varbinary_field = build_binary_field



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import os

import sqlalchemy as sa

from django.conf import settings


DIALECT_MAP = {
    "django.db.backends.sqlite3": "sqlite",
    "django.db.backends.postgresql": "postgresql",
    "django.db.backends.mysql": "mysql",
    "django.db.backends.oracle": "oracle",
    "sqlserver_ado": "mssql",
}


def get_settings(alias):
    if hasattr(settings, "SQLALCHEMY_CONNECTIONS") and alias in settings.SQLALCHEMY_CONNECTIONS:
        return settings.SQLALCHEMY_CONNECTIONS[alias]

    return settings.DATABASES[alias]


def make_url(alias_or_url):
    """
    Generates a URL either from environment, DATABASES or SQLALCHEMY_CONNECTIONS settings
    ---------------------------------------------------------------------
    alias_or_url: str
        name of the alias or url as string
    """
    try:
        return sa.engine.url.make_url(alias_or_url), {}

    except sa.exc.ArgumentError:
        pass

    alias = alias_or_url

    url = sa.engine.url.make_url(os.environ.get(alias.upper() + "_URL", None))
    if url:
        return url, {}

    return make_url_from_settings(alias)


def make_url_from_settings(alias):
    """
    Generates a URL using the alias in settings
    -------------------------------------------
    alias: str
        name of the alias
    """

    data = get_settings(alias)

    if "DIALECT" not in data:
        data["DIALECT"] = DIALECT_MAP.get(data["ENGINE"]) or data["ENGINE"].split(".")[-1]

    names = [data["DIALECT"].lower()]

    if "DRIVER" in data:
        names.append(data["DRIVER"].lower())

    drivername = "+".join(names)

    url = sa.engine.url.URL(drivername, username=data.get("USER") or None, password=data.get("PASSWORD") or None)

    url.host = data.get("HOST") or None
    url.database = data.get("NAME") or None
    try:
        url.port = int(data.get("PORT"))
    except Exception:
        pass

    url.query.update(data.get("QUERY", {}))

    return url, data.get("OPTIONS", {})



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

from .base import GenericViewSet, ModelViewSet, ReadOnlyModelViewSet  # noqa
from .mixins import (  # noqa
    CreateModelMixin, DeleteModelMixin, ListModelMixin, ModelFormMixin, RetrieveModelMixin, UpdateModelMixin
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import sqlalchemy as sa

from django_sorcery.db import meta  # noqa

from ..base import TestCase
from ..models import COLORS, Owner, Vehicle, VehicleType, Vertex, AllKindsOfFields


class TestModelMeta(TestCase):

    def test_model_meta(self):
        info = meta.model_info(Owner)

        self.assertEqual(repr(info), "<model_info(Owner)>")
        self.assertListEqual(list(info.primary_keys.keys()), ["id"])
        self.assertEqual(info.primary_keys["id"].property, Owner.id.property)

        self.assertListEqual(list(info.properties.keys()), ["first_name", "last_name"])
        self.assertEqual(info.properties["first_name"].property, Owner.first_name.property)
        self.assertEqual(info.properties["last_name"].property, Owner.last_name.property)

        self.assertListEqual(list(info.relationships.keys()), ["vehicles"])
        self.assertTrue(info.relationships["vehicles"].relationship is Owner.vehicles.property)

        self.assertListEqual(info.field_names, ["id", "first_name", "last_name", "vehicles"])

    def test_model_meta_with_mapper(self):
        mapper = Vehicle.owner.property.parent
        self.assertEqual(meta.model_info(mapper), meta.model_info(Vehicle))


class TestCompositeMeta(TestCase):

    def test_composite_meta(self):

        info = meta.model_info(Vertex)
        self.assertListEqual(list(info.composites.keys()), ["start", "end"])

        start = info.composites["start"]
        self.assertListEqual(sorted(start.properties.keys()), ["x", "y"])
        self.assertEqual(start.properties["x"].property, Vertex.x1.property)

        self.assertListEqual(sorted(start.field_names), ["x", "y"])


class TestRelationshipMeta(TestCase):

    def test_relationship_meta(self):
        info = meta.model_info(Owner)

        rel = info.relationships["vehicles"]

        self.assertEqual(rel.related_model, Vehicle)
        self.assertEqual(rel.name, "vehicles")
        self.assertEqual(rel.direction, sa.orm.relationships.ONETOMANY)
        self.assertEqual(list(rel.foreign_keys), Vehicle._owner_id.property.columns)
        self.assertTrue(rel.uselist)
        self.assertEqual(repr(rel), "<relation_info(Owner.vehicles)>")


class TestColumnMeta(TestCase):

    def test_columnInfo(self):
        info = meta.model_info(Vehicle)

        col = info.primary_keys["id"]
        self.assertEqual(col.name, "id")
        self.assertDictEqual({"label": "Id", "help_text": "The primary key", "required": True}, col.field_kwargs)

        col = info.properties["type"]
        self.assertDictEqual(
            {"enum_class": VehicleType, "help_text": None, "label": "Type", "max_length": 3, "required": True},
            col.field_kwargs,
        )

        col = info.properties["paint"]
        self.assertDictEqual(
            {"choices": COLORS, "help_text": None, "label": "Paint", "max_length": 6, "required": False},
            col.field_kwargs,
        )
        self.assertEqual(col.parent_model, Vehicle)
        self.assertEqual(repr(col), "<column_info(Vehicle.paint)>")

        info = meta.model_info(AllKindsOfFields)
        col = info.properties["decimal"]
        self.assertDictEqual(
            {"decimal_places": None, "help_text": None, "label": "Decimal", "max_digits": None, "required": False},
            col.field_kwargs,
        )



# -*- coding: utf-8 -*-
"""
Helper functions for creating Form classes from SQLAlchemy models.
"""
from __future__ import absolute_import, print_function, unicode_literals

import six

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.forms.forms import BaseForm, DeclarativeFieldsMetaclass
from django.forms.models import ModelFormOptions
from django.forms.utils import ErrorList
from django.utils.module_loading import import_string

from .db.models import model_to_dict
from .field_mapping import ALL_FIELDS, ModelFieldMapper, apply_limit_choices_to_form_field


class SQLAModelFormOptions(ModelFormOptions):

    def __init__(self, options=None):
        super(SQLAModelFormOptions, self).__init__(options=options)
        self.session = getattr(options, "session", None)


class ModelFormMetaclass(DeclarativeFieldsMetaclass):

    def __new__(cls, name, bases, attrs):
        mcs = super(ModelFormMetaclass, cls).__new__(cls, name, bases, attrs)

        base_formfield_callback = None
        for base in bases:
            if hasattr(base, "Meta") and hasattr(base.Meta, "formfield_callback"):
                base_formfield_callback = base.Meta.formfield_callback
                break

        formfield_callback = attrs.pop("formfield_callback", base_formfield_callback)

        if bases == (BaseModelForm,):
            return mcs

        opts = mcs._meta = SQLAModelFormOptions(getattr(mcs, "Meta", None))

        for opt in ["fields", "exclude", "localized_fields"]:
            value = getattr(opts, opt)
            if isinstance(value, six.string_types) and value != ALL_FIELDS:
                raise TypeError(
                    "%(model)s.Meta.%(opt)s cannot be a string. Did you mean to type: ('%(value)s',)?"
                    % {"model": mcs.__name__, "opt": opt, "value": value}
                )

        if opts.model:
            if opts.fields is None and opts.exclude is None:
                raise ImproperlyConfigured(
                    "Creating a ModelForm without either the 'fields' attribute or the 'exclude' attribute is "
                    "prohibited; form %s needs updating." % name
                )

            if opts.fields == ALL_FIELDS:
                opts.fields = None

            mapper = import_string(settings.SQLALCHEMY_FORM_MAPPER) if hasattr(
                settings, "SQLALCHEMY_FORM_MAPPER"
            ) else ModelFieldMapper

            mcs.base_fields = mapper(opts, formfield_callback, apply_limit_choices_to=False).get_fields()

        else:
            mcs.base_fields = mcs.declared_fields

        return mcs


class BaseModelForm(BaseForm):

    def __init__(
        self,
        data=None,
        files=None,
        auto_id="id_%s",
        prefix=None,
        initial=None,
        error_class=ErrorList,
        label_suffix=None,
        empty_permitted=False,
        instance=None,
        use_required_attribute=None,
        renderer=None,
        session=None,
    ):
        opts = self._meta
        opts.session = opts.session or session
        if opts.model is None:
            raise ValueError("ModelForm has no model class specified.")

        if opts.session is None:
            raise ValueError("ModelForm has no session specified.")

        self.instance = opts.model() if instance is None else instance
        object_data = self.model_to_dict()
        object_data.update(initial or {})
        self._validate_unique = False
        super(BaseModelForm, self).__init__(
            data,
            files,
            auto_id,
            prefix,
            object_data,
            error_class,
            label_suffix,
            empty_permitted,
            use_required_attribute=use_required_attribute,
            renderer=renderer,
        )
        for field in self.fields.values():
            apply_limit_choices_to_form_field(field)

    def model_to_dict(self):
        opts = self._meta
        return model_to_dict(self.instance, opts.fields, opts.exclude)

    def update_attribute(self, instance, name, field, value):
        field_setter = getattr(self, "set_" + name, None)
        if field_setter:
            field_setter(instance, name, field, value)
        else:
            setattr(instance, name, value)

    def save(self, flush=True, **kwargs):
        opts = self._meta

        if self.errors:
            raise ValueError(
                "The %s could not be saved because the data didn't validate." % (self.instance.__class__.__name__,)
            )

        self.instance = self.save_instance()

        if flush:
            opts.session.flush()

        return self.instance

    def save_instance(self, instance=None, cleaned_data=None):

        instance = instance or self.instance
        cleaned_data = cleaned_data or self.cleaned_data

        for name, field in self.fields.items():
            if name in cleaned_data:
                self.update_attribute(self.instance, name, field, cleaned_data[name])

        if instance not in self._meta.session:
            self._meta.session.add(instance)
        return instance


class ModelForm(six.with_metaclass(ModelFormMetaclass, BaseModelForm)):
    pass


def modelform_factory(model, form=ModelForm, formfield_callback=None, **kwargs):
    defaults = [
        "fields",
        "exclude",
        "widgets",
        "localized_fields",
        "labels",
        "help_texts",
        "error_messages",
        "field_classes",
        "session",
    ]

    attrs = {"model": model}
    for key in defaults:
        value = kwargs.get(key)
        if value is not None:
            attrs[key] = value

    bases = (form.Meta,) if hasattr(form, "Meta") else tuple()
    meta = type(str("Meta"), bases, attrs)
    if formfield_callback:
        meta.formfield_callback = staticmethod(formfield_callback)

    class_name = model.__name__ + "Form"

    if getattr(meta, "fields", None) is None and getattr(meta, "exclude", None) is None:
        raise ImproperlyConfigured(
            "Calling modelform_factory without defining 'fields' or 'exclude' explicitly is prohibited."
        )

    return type(form)(str(class_name), (form,), {"Meta": meta, "formfield_callback": formfield_callback})



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from sqlalchemy import event, orm

from . import signals


class SignallingSession(orm.Session):

    def __init__(self, *args, **kwargs):
        super(SignallingSession, self).__init__(*args, **kwargs)

        @event.listens_for(self, "before_flush")
        def before_flush(session, flush_context, instances):
            signals.before_flush.send(session, flush_context=flush_context, instances=instances)

        @event.listens_for(self, "after_flush")
        def after_flush(session, flush_context):
            signals.after_flush.send(session, flush_context=flush_context)

    def query(self, *args, **kwargs):
        """
        Override to try to use the model.query_class
        """
        if len(args) == 1 and hasattr(args[0], "query_class") and args[0].query_class is not None:
            return args[0].query_class(*args, session=self, **kwargs)

        return super(SignallingSession, self).query(*args, **kwargs)

    def commit(self):
        is_main = self.transaction and (self.transaction._parent is None or not self.transaction.nested)

        if is_main:
            signals.before_commit.send(self)
            signals.before_scoped_commit.send(self)

        super(SignallingSession, self).commit()

        if is_main:
            signals.after_scoped_commit.send(self)
            signals.after_commit.send(self)

    def rollback(self):
        super(SignallingSession, self).rollback()

        if self.transaction and (self.transaction._parent is None or self.transaction.nested):
            signals.after_scoped_rollback.send(self)
            signals.after_rollback.send(self)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django_sorcery.db import signals  # noqa

from ..base import TestCase


class TestSession(TestCase):

    def test_session_signals(self):
        pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import os


def absjoin(*args):
    return os.path.normpath(os.path.abspath(os.path.join(*args)))


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

PROJECT_ROOT = os.path.dirname(absjoin(__file__))
STATIC_ROOT = absjoin(PROJECT_ROOT, "..", "static")

DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": os.path.join(BASE_DIR, "db.sqlite3")},
    "fromdbs": {"ENGINE": "django.db.backends.sqlite3", "NAME": os.path.join(BASE_DIR, "fromdbs.sqlite3")},
}

SQLALCHEMY_CONNECTIONS = {
    "default": {"DIALECT": "sqlite"},
    "test": {"DIALECT": "sqlite"},
    "minimal": {"DIALECT": "sqlite"},
    "minimal_backpop": {"DIALECT": "sqlite"},
}

INSTALLED_APPS = ["tests.apps.TestConfig", "django.contrib.staticfiles"]

ROOT_URLCONF = "tests.urls"

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

SECRET_KEY = "secret"

ALLOWED_HOSTS = ["*"]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(os.path.dirname(__file__), "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug", "django.template.context_processors.request"
            ]
        },
    }
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django import test

from .models import db


try:
    from unittest import mock  # noqa
except ImportError:
    import mock  # noqa


class TestCase(test.TestCase):

    def setUp(self):
        super(TestCase, self).setUp()
        self.factory = test.RequestFactory()

    def tearDown(self):
        db.rollback()
        db.remove()
        super(TestCase, self).tearDown()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from bs4 import BeautifulSoup

from django.core.exceptions import ImproperlyConfigured

from django_sorcery.forms import ALL_FIELDS, ModelForm, modelform_factory

from .base import TestCase
from .models import Option, Owner, Vehicle, VehicleType, db


class TestModelForm(TestCase):

    def setUp(cls):
        super(TestModelForm, cls).setUp()
        db.add(Owner(id=1, first_name="Test", last_name="Owner"))
        db.add_all(
            [
                Option(id=1, name="Option 1"),
                Option(id=2, name="Option 2"),
                Option(id=3, name="Option 3"),
                Option(id=4, name="Option 4"),
            ]
        )
        db.flush()

    def test_modelform_factory_fields(self):
        form_class = modelform_factory(Vehicle, fields=ALL_FIELDS, session=db)
        form = form_class()
        self.assertListEqual(
            sorted(form.fields.keys()), ["created_at", "is_used", "name", "options", "owner", "paint", "parts", "type"]
        )

    def test_modelform_factory_instance_validate(self):
        vehicle = Vehicle(owner=Owner.query.get(1))
        form_class = modelform_factory(Vehicle, fields=ALL_FIELDS, session=db)
        form = form_class(instance=vehicle, data={"name": "testing"})
        self.assertFalse(form.is_valid())
        self.assertEqual({"type": ["This field is required."], "owner": ["This field is required."]}, form.errors)

    def test_modelform_factory_instance_save(self):
        form_class = modelform_factory(Vehicle, fields=ALL_FIELDS, session=db)
        data = {"name": "testing", "type": "car", "owner": 1}
        form = form_class(data=data)
        self.assertTrue(form.is_valid(), form.errors)
        form.save()

    def test_modelform_factory_modelchoicefield_choices(self):
        form_class = modelform_factory(Vehicle, fields=ALL_FIELDS, session=db)
        data = {"name": "testing", "type": "car", "owner": 1}
        form = form_class(data=data)

        owner_choices = form.fields["owner"].choices
        option_choices = form.fields["options"].choices
        self.assertEqual(len(owner_choices), 2)
        self.assertEqual(len(option_choices), 4)

    def test_modelform_factory_new_render(self):
        form_class = modelform_factory(Vehicle, fields=ALL_FIELDS, session=db)
        form = form_class(data={})

        self.assertTrue(form.is_bound)
        self.assertEqual(form.errors, {"owner": ["This field is required."], "type": ["This field is required."]})
        self.assertEqual(form.initial, {"paint": None, "created_at": None, "type": None, "name": None, "is_used": None})
        self.assertEqual(
            form.cleaned_data,
            {"paint": "", "created_at": None, "options": [], "parts": [], "name": u"", "is_used": None},
        )

        form.order_fields(sorted(form.fields.keys()))

        soup = BeautifulSoup(form.as_p(), "html.parser")
        expected_soup = BeautifulSoup(
            "".join(
                [
                    "<p>",
                    '  <label for="id_created_at">Created at:</label>',
                    '  <input type="text" name="created_at" id="id_created_at" />' "</p>",
                    "<p>",
                    '  <label for="id_is_used">Is used:</label>',
                    '  <select name="is_used" id="id_is_used">',
                    '    <option value="1" selected>Unknown</option>',
                    '    <option value="2">Yes</option>',
                    '    <option value="3">No</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_name">Name:</label>',
                    '  <input id="id_name" name="name" type="text" />',
                    '  <span class="helptext">' "    The name of the vehicle",
                    "  </span>",
                    "</p>",
                    "<p>",
                    '  <label for="id_options">Options:</label>',
                    '  <select id="id_options" multiple="multiple" name="options">',
                    "    <option value=\"1\">Option(id=1, name='Option 1')</option>",
                    "    <option value=\"2\">Option(id=2, name='Option 2')</option>",
                    "    <option value=\"3\">Option(id=3, name='Option 3')</option>",
                    "    <option value=\"4\">Option(id=4, name='Option 4')</option>",
                    "  </select>",
                    "</p>",
                    '<ul class="errorlist">',
                    "  <li>",
                    "    This field is required.",
                    "  </li>",
                    "</ul>",
                    "<p>",
                    '  <label for="id_owner">Owner:</label>',
                    '  <select id="id_owner" name="owner" required>',
                    "    <option selected value>---------</option>",
                    "    <option value=\"1\">Owner(id=1, first_name='Test', last_name='Owner')</option>",
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_paint">Paint:</label>',
                    '  <select id="id_paint" name="paint">',
                    "    <option selected value></option>",
                    '    <option value="red">red</option>',
                    '    <option value="green">green</option>',
                    '    <option value="blue">blue</option>',
                    '    <option value="silver">silver</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_parts">Parts:</label>',
                    '  <select id="id_parts" multiple="multiple" name="parts">',
                    "  </select>",
                    "</p>",
                    '<ul class="errorlist">',
                    "  <li>",
                    "    This field is required.",
                    "  </li>",
                    "</ul>",
                    "<p>",
                    '  <label for="id_type">Type:</label>',
                    '  <select id="id_type" name="type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                ]
            ),
            "html.parser",
        )

        self.assertEqual(soup.prettify(), expected_soup.prettify())

    def test_modelform_factory_instance_render(self):
        form_class = modelform_factory(Vehicle, fields=ALL_FIELDS, session=db)
        vehicle = Vehicle(owner=Owner.query.get(1), type=VehicleType.car)
        form = form_class(instance=vehicle, data={})

        self.assertTrue(form.is_bound)
        self.assertEqual(form.errors, {"owner": ["This field is required."], "type": ["This field is required."]})
        self.assertEqual(
            form.initial,
            {"paint": None, "created_at": None, "type": VehicleType.car, "name": None, "is_used": None, "owner": 1},
        )
        self.assertEqual(
            form.cleaned_data,
            {"paint": "", "created_at": None, "options": [], "parts": [], "name": u"", "is_used": None},
        )

        form.order_fields(sorted(form.fields.keys()))

        soup = BeautifulSoup(form.as_p(), "html.parser")
        expected_soup = BeautifulSoup(
            "".join(
                [
                    "<p>",
                    '  <label for="id_created_at">Created at:</label>',
                    '  <input type="text" name="created_at" id="id_created_at" />' "</p>",
                    "<p>",
                    '  <label for="id_is_used">Is used:</label>',
                    '  <select name="is_used" id="id_is_used">',
                    '    <option value="1" selected>Unknown</option>',
                    '    <option value="2">Yes</option>',
                    '    <option value="3">No</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_name">Name:</label>',
                    '  <input id="id_name" name="name" type="text" />',
                    '  <span class="helptext">' "    The name of the vehicle",
                    "  </span>",
                    "</p>",
                    "<p>",
                    '  <label for="id_options">Options:</label>',
                    '  <select id="id_options" multiple="multiple" name="options">',
                    "    <option value=\"1\">Option(id=1, name='Option 1')</option>",
                    "    <option value=\"2\">Option(id=2, name='Option 2')</option>",
                    "    <option value=\"3\">Option(id=3, name='Option 3')</option>",
                    "    <option value=\"4\">Option(id=4, name='Option 4')</option>",
                    "  </select>",
                    "</p>",
                    '<ul class="errorlist">',
                    "  <li>",
                    "    This field is required.",
                    "  </li>",
                    "</ul>",
                    "<p>",
                    '  <label for="id_owner">Owner:</label>',
                    '  <select id="id_owner" name="owner" required>',
                    "    <option selected value>---------</option>",
                    "    <option value=\"1\">Owner(id=1, first_name='Test', last_name='Owner')</option>",
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_paint">Paint:</label>',
                    '  <select id="id_paint" name="paint">',
                    "    <option selected value></option>",
                    '    <option value="red">red</option>',
                    '    <option value="green">green</option>',
                    '    <option value="blue">blue</option>',
                    '    <option value="silver">silver</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_parts">Parts:</label>',
                    '  <select id="id_parts" multiple="multiple" name="parts">',
                    "  </select>",
                    "</p>",
                    '<ul class="errorlist">',
                    "  <li>",
                    "    This field is required.",
                    "  </li>",
                    "</ul>",
                    "<p>",
                    '  <label for="id_type">Type:</label>',
                    '  <select id="id_type" name="type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                ]
            ),
            "html.parser",
        )

        self.maxDiff = None
        self.assertHTMLEqual(soup.prettify(), expected_soup.prettify())

    def test_form_field_callback_in_base_meta(self):

        self.callback_called = False

        def callback(*args, **kwargs):
            self.callback_called = True

        class OwnerBaseForm(ModelForm):

            class Meta:
                model = Owner
                session = db
                fields = ALL_FIELDS
                formfield_callback = staticmethod(callback)

        class OwnerForm(OwnerBaseForm):
            pass

        self.assertTrue(self.callback_called)

    def test_fields_bad_value(self):

        with self.assertRaises(TypeError) as ctx:
            modelform_factory(Owner, ModelForm, fields="abc1234")

        self.assertEqual(
            ctx.exception.args, ("OwnerForm.Meta.fields cannot be a string. Did you mean to type: ('abc1234',)?",)
        )

    def test_empty_fields_and_exclude(self):

        with self.assertRaises(ImproperlyConfigured) as ctx:

            class OwnerForm(ModelForm):

                class Meta:
                    model = Owner

        self.assertEqual(
            ctx.exception.args,
            (
                "Creating a ModelForm without either the 'fields' attribute or the 'exclude' attribute is prohibited; "
                "form OwnerForm needs updating.",
            ),
        )

    def test_modelform_no_model(self):

        class OwnerForm(ModelForm):
            pass

        with self.assertRaises(ValueError) as ctx:
            OwnerForm()

        self.assertEqual(ctx.exception.args, ("ModelForm has no model class specified.",))

    def test_modelform_no_session(self):

        class OwnerForm(ModelForm):

            class Meta:
                model = Owner
                fields = ALL_FIELDS

        with self.assertRaises(ValueError) as ctx:
            OwnerForm()

        self.assertEqual(ctx.exception.args, ("ModelForm has no session specified.",))

    def test_modelform_custom_setter(self):

        class OwnerForm(ModelForm):

            class Meta:
                model = Owner
                session = db
                fields = ("first_name",)

            def set_first_name(self, instance, name, field, value):
                instance.first_name = "other"

        form = OwnerForm(data={"first_name": "something"})

        instance = form.save()

        self.assertEqual(instance.first_name, "other")

    def test_modelform_save_with_errors(self):

        class VehicleForm(ModelForm):

            class Meta:
                model = Vehicle
                session = db
                fields = ("type", "name")

        form = VehicleForm(data={"name": "something"})

        self.assertFalse(form.is_valid())
        with self.assertRaises(ValueError) as ctx:
            form.save()

        self.assertEqual(ctx.exception.args, ("The Vehicle could not be saved because the data didn't validate.",))

    def test_modelform_factory_with_formfield_callback(self):

        self.callback_called = False

        def callback(*args, **kwargs):
            self.callback_called = True

        modelform_factory(Owner, fields=ALL_FIELDS, formfield_callback=callback, session=db)

        self.assertTrue(self.callback_called)

    def test_modelform_factory_with_no_fields_exclude(self):

        with self.assertRaises(ImproperlyConfigured) as ctx:
            modelform_factory(Owner, session=db)

        self.assertEqual(
            ctx.exception.args,
            ("Calling modelform_factory without defining 'fields' or 'exclude' explicitly is prohibited.",),
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.test import TestCase

from django_sorcery import compat


class TestCompat(TestCase):

    def test_suppress(self):

        with compat.suppress(AssertionError):
            assert 1 == 2



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.http import Http404
from django_sorcery import shortcuts

from .models import Owner, db
from .base import TestCase


class TestShortcuts(TestCase):

    def setUp(self):
        super(TestShortcuts, self).setUp()
        db.add_all(
            [
                Owner(first_name="Test 1", last_name="Owner 1"),
                Owner(first_name="Test 2", last_name="Owner 2"),
                Owner(first_name="Test 3", last_name="Owner 3"),
                Owner(first_name="Test 4", last_name="Owner 4"),
            ]
        )
        db.flush()

    def test_get_object_or_404(self):

        obj = shortcuts.get_object_or_404(Owner, id=1)
        self.assertIsNotNone(obj)

        with self.assertRaises(Http404):
            shortcuts.get_object_or_404(Owner, id=999)

        obj = shortcuts.get_object_or_404(Owner.query, id=1)
        self.assertIsNotNone(obj)

        with self.assertRaises(Http404):
            shortcuts.get_object_or_404(Owner.query, id=999)

    def test_get_list_or_404(self):

        obj = shortcuts.get_list_or_404(Owner)
        self.assertTrue(list(obj))

        with self.assertRaises(Http404):
            shortcuts.get_list_or_404(Owner, Owner.first_name.ilike("XXX"))

        obj = shortcuts.get_list_or_404(Owner, first_name="Test 1")
        self.assertTrue(list(obj))

        with self.assertRaises(Http404):
            shortcuts.get_list_or_404(Owner, first_name="XXX")
